@@ -134,6 +134,8 @@ static struct
     {0,		    0, 0}
 };
 
+#if 0
+    /* not used? */
     static int
 hex_digit(int c)
 {
@@ -144,6 +146,7 @@ hex_digit(int c)
 	return c - 'a' + 10;
     return -1000;
 }
+#endif
 
 static int characterWidth = -1;
 static int characterHeight = -1;
@@ -376,10 +379,19 @@ D("Select UP detected\n");
 				break;
 			case IDCMP_VANILLAKEY:
 			{
-				char_u string[1];
+				char_u string[3];
+
 				string[0] = (char_u)code;
+				if (code == CSI)
+				{
+				    /* Insert CSI as K_CSI.  Untested! */
+				    string[1] = KS_EXTRA;
+				    string[2] = KE_CSI;
+				    add_to_input_buf(string, 3);
+				}
+				else
+				    add_to_input_buf(string, 1);
 				returnEvent = ev_KeyStroke;
-				add_to_input_buf(string,1);
 				break;
 			case IDCMP_RAWKEY:
 				if (msg->Qualifier & IEQUALIFIER_LSHIFT)
@@ -553,7 +565,7 @@ charEventHandler(int wtime)
  * add primary menu
  */
     void
-gui_mch_add_menu_item(GuiMenu *menu, GuiMenu *parent, int idx)
+gui_mch_add_menu_item(VimMenu *menu, VimMenu *parent, int idx)
 {
     union myMenuItemUnion *menuItemUnion = NULL;
     struct IntuiText *menutext = NULL;
@@ -562,7 +574,7 @@ gui_mch_add_menu_item(GuiMenu *menu, GuiMenu *parent, int idx)
     assert(parent != NULL);
 
     /* Don't add menu separator */
-    if (gui_is_separator(menu->name))
+    if (is_menu_separator(menu->name))
 	return;
 
     if (parent->menuItemPtr == NULL)
@@ -653,12 +665,12 @@ getMenu(struct RastPort *rast, int left, STRPTR name)
  * add  1st level submenu item
  */
     void
-gui_mch_add_menu(GuiMenu *menu, GuiMenu *parent, int idx)
+gui_mch_add_menu(VimMenu *menu, VimMenu *parent, int idx)
 {
     struct Menu	*newMenu;
     int		pos = 0;
 
-    if (!gui_menubar_menu(menu->name))
+    if (!menubar_menu(menu->name))
 	return;
 
     menu->menuPtr = newMenu = getMenu(gui.window->RPort, 0, menu->dname);
@@ -731,26 +743,35 @@ atexitDoThis(void)
     gui_mch_exit(-1);
 }
 
+/*
+ * Check if the GUI can be started.  Called before gvimrc is sourced.
+ * Return OK or FAIL.
+ */
+    int
+gui_mch_init_check(void)
+{
+    if (execBase && gfxBase && layersBase)
+	return OK;
+    return FAIL;
+}
+
     int
 gui_mch_init(void)
 {
     int returnCode = FAIL; /* assume failure*/
 
-    if (execBase && gfxBase && layersBase)
+    gui.window = OpenWindowTagList(&vimNewWindow, tags);
+    if (gui.window)
     {
-
-	gui.window = OpenWindowTagList(&vimNewWindow, tags);
-	if (gui.window)
-	{
-	    gui.in_use = TRUE;
-	    gui.in_focus=TRUE;
-	    SetDrMd(gui.window->RPort, JAM2);
-	    atexit(atexitDoThis);
-	    TextDimensions();
-	    returnCode = OK; /* we've had sucess */
-	}
-	gui.menu = NULL;
+	gui.in_use = TRUE;
+	gui.in_focus=TRUE;
+	SetDrMd(gui.window->RPort, JAM2);
+	atexit(atexitDoThis);
+	TextDimensions();
+	returnCode = OK; /* we've had sucess */
     }
+    gui.menu = NULL;
+
     return returnCode;
 }
 
@@ -784,6 +805,26 @@ gui_mch_exit(int returnCode)
     }
 }
 
+/*
+ * Get the position of the top left corner of the window.
+ */
+    int
+gui_mch_get_winpos(int *x, int *y)
+{
+    /* TODO */
+    return FAIL;
+}
+
+/*
+ * Set the position of the top left corner of the window to the given
+ * coordinates.
+ */
+    void
+gui_mch_set_winpos(int x, int y)
+{
+    /* TODO */
+}
+
     void
 gui_mch_set_winsize(int width, int height, int min_width, int min_height,
 		    int base_width, int base_height)
@@ -859,11 +900,13 @@ gui_mch_set_font(GuiFont font)
     /*D("gui_mch_set_font");*/
 }
 
+#if 0 /* not used */
     int
 gui_mch_same_font(GuiFont f1, GuiFont f2)
 {
     D("gui_mch_same_font");
 }
+#endif
 
     void
 gui_mch_free_font(GuiFont font)
@@ -1027,7 +1070,7 @@ gui_mch_beep(void)
 }
 
     void
-gui_mch_flash(void)
+gui_mch_flash(int msec)
 {
     D("gui_mch_flash");
 }
@@ -1174,20 +1217,20 @@ gui_mch_set_menu_pos(int x, int y, int w, int h)
 }
 
     void
-gui_mch_destroy_menu(GuiMenu *menu)
+gui_mch_destroy_menu(VimMenu *menu)
 {
     D("gui_mch_destroy_menu");
     ClearMenuStrip(gui.window);
 }
 
     void
-gui_mch_menu_grey(GuiMenu *menu, int grey)
+gui_mch_menu_grey(VimMenu *menu, int grey)
 {
     D("gui_mch_menu_grey");
 }
 
     void
-gui_mch_menu_hidden(GuiMenu *menu, int hidden)
+gui_mch_menu_hidden(VimMenu *menu, int hidden)
 {
     D("gui_mch_menu_hidden");
     ClearMenuStrip(gui.window);
@@ -1257,7 +1300,7 @@ gui_mch_setmouse(x, y)
 }
 
     void
-gui_mch_show_popupmenu(GuiMenu *menu)
+gui_mch_show_popupmenu(VimMenu *menu)
 {
     /* TODO */
 }
@@ -1293,6 +1336,7 @@ gui_mch_get_lightness(GuiColor pixel)
     return (int)rc;
 }
 
+#if (defined(SYNTAX_HL) && defined(WANT_EVAL)) || defined(PROTO)
     char_u *
 gui_mch_get_rgb(GuiColor pixel)
 {
@@ -1305,5 +1349,6 @@ gui_mch_get_rgb(GuiColor pixel)
 
     return retval;
 }
+#endif
 
 #endif /* USE_AMIGA_GUI*/
