@@ -42,6 +42,9 @@ static int cbrk_pressed = FALSE;    /* set by ctrl-break interrupt */
 static int ctrlc_pressed = FALSE;   /* set when ctrl-C or ctrl-break detected */
 static int delayed_redraw = FALSE;  /* set when ctrl-C detected */
 
+static int bioskey_read = _NKEYBRD_READ;   /* bioskey() argument: read key */
+static int bioskey_ready = _NKEYBRD_READY; /* bioskey() argument: key ready? */
+
 #ifdef USE_MOUSE
 static int mouse_avail = FALSE;		/* mouse present */
 static int mouse_active;		/* mouse enabled */
@@ -73,7 +76,7 @@ unsigned long	S_ulScreenBase = 0xb8000;
 unsigned short	S_uiAttribute = 0;
 int		S_iCurrentRow = 0;	/* These are 0 offset */
 int		S_iCurrentColumn = 0;
-int		S_iLeft = 0;	/* These are 1 offset */
+int		S_iLeft = 0;	/* Scroll window; these are 1 offset */
 int		S_iTop = 0;
 int		S_iRight = 0;
 int		S_iBottom = 0;
@@ -87,6 +90,9 @@ mygotoxy(int x, int y)
     gotoxy(x,y);		/* set cursor position */
 }
 
+/*
+ * Scroll the screen one line up, clear the last line.
+ */
     static void
 myscroll(void)
 {
@@ -282,6 +288,7 @@ translate_altkeys(int rawkey)
     }
     return rawkey;
 }
+
 /*
  * Set normal fg/bg color, based on T_ME.  Called whem t_me has been set.
  */
@@ -592,7 +599,8 @@ WaitForChar(long msec)
 	}
 #endif
 
-	if ((p_consk ? cons_kbhit() : p_biosk ? bioskey(1) : kbhit())
+	if ((p_consk ? cons_kbhit()
+				 : p_biosk ? bioskey(bioskey_ready) : kbhit())
 		|| cbrk_pressed
 #ifdef USE_MOUSE
 						    || mouse_click >= 0
@@ -604,9 +612,15 @@ WaitForChar(long msec)
 	 * We busy-wait here.  Unfortunately, delay() and usleep() have been
 	 * reported to give problems with the original Windows 95.  This is
 	 * fixed in service pack 1, but not everybody installed that.
+	 * The DJGPP implementation of usleep() uses a busy-wait loop too.
 	 */
 	if (msec != FOREVER && biostime(0, 0L) > starttime + msec / BIOSTICK)
 	    break;
+
+#ifdef DJGPP
+	/* Yield the CPU to the next process. */
+	__dpmi_yield();
+#endif
     }
     return FALSE;
 }
@@ -870,22 +884,27 @@ mch_inchar(
 #endif
 	if (p_biosk && !p_consk)
 	{
-	    while ((len == 0 || bioskey(1)) && len < maxlen)
+	    while ((len == 0 || bioskey(bioskey_ready)) && len < maxlen)
 	    {
-		c = translate_altkeys(bioskey(0));	/* get the key */
+		c = translate_altkeys(bioskey(bioskey_read)); /* get the key */
 		/*
 		 * translate a few things for inchar():
 		 * 0x0000 == CTRL-break		-> 3	(CTRL-C)
 		 * 0x0300 == CTRL-@		-> NUL
 		 * 0xnn00 == extended key code	-> K_NUL, nn
+		 * 0xnne0 == enhanced keyboard	-> K_NUL, nn
 		 * K_NUL			-> K_NUL, 3
 		 */
 		if (c == 0)
 		    c = 3;
 		else if (c == 0x0300)
 		    c = NUL;
-		else if ((c & 0xff) == 0 || c == K_NUL
-				 || c == 0x4e2b || c == 0x4a2d || c == 0x372a)
+		else if ((c & 0xff) == 0
+			|| c == K_NUL
+			|| c == 0x4e2b
+			|| c == 0x4a2d
+			|| c == 0x372a
+			|| ((c & 0xff) == 0xe0 && c != 0xe0))
 		{
 		    if (c == K_NUL)
 			c = 3;
@@ -990,6 +1009,7 @@ catch_cbrk(void)
     ctrlc_pressed = TRUE;
 }
 
+#ifndef DJGPP
 /*
  * ctrl-break handler for DOS. Never called when a ctrl-break is typed, because
  * we catch interrupt 1b. If you type ctrl-C while Vim is waiting for a
@@ -1018,6 +1038,7 @@ catch_cint(bp, di, si, ds, es, dx, cx, bx, ax)
     if (_osmajor >= 3)
 	ax |= 3;	    /* set AL to 3 */
 }
+#endif
 
 /*
  * Set the interrupt vectors for use with Vim on or off.
@@ -1133,6 +1154,32 @@ mch_windinit(void)
     regs.h.bh = 0x00;
     int86(0x10, &regs, &regs);
 
+    /*
+     * Test if we have an enhanced AT keyboard.  Write 0xFFFF to the keyboard
+     * buffer and try to read it back.  If we can't in 16 tries, it's an old
+     * type XT keyboard.
+     */
+    regs.h.ah = 0x05;
+    regs.x.cx = 0xffff;
+    int86(0x16, &regs, &regs);
+    if (regs.h.al != 1)	/* skip this when keyboard buffer is full */
+    {
+	int i;
+
+	for (i = 0; i < 16; ++i)
+	{
+	    regs.h.ah = 0x10;
+	    int86(0x16, &regs, &regs);
+	    if (regs.x.ax == 0xffff)
+		break;
+	}
+	if (i == 16)	/* 0xffff not read, must be old keyboard */
+	{
+	    bioskey_read = 0;
+	    bioskey_ready = 1;
+	}
+    }
+
 #ifdef MCH_CURSOR_SHAPE
     /* Save the old cursor shape */
     mch_restore_cursor_shape(FALSE);
@@ -1638,16 +1685,36 @@ mch_get_winsize(void)
     return OK;
 }
 
+#if defined(DJGPP) || defined(PROTO)
+/*
+ * Check the number of Columns with a BIOS call.  This avoids a crash of the
+ * DOS console when 'columns' is set to a too large value.
+ */
+    void
+mch_check_columns()
+{
+    static union REGS	regs;
+
+    regs.h.ah = 0x0f;
+    (void)int86(0x10, &regs, &regs);
+    if ((unsigned)Columns > (unsigned)regs.h.ah)
+	Columns = (unsigned)regs.h.ah;
+}
+#endif
+
 /*
  * Set the active window for delline/insline.
  */
     void
 set_window(void)
 {
+    if (term_console)
+    {
 #ifndef DJGPP
-    _video.screenheight = Rows;
+	_video.screenheight = Rows;
 #endif
-    mywindow(1, 1, Columns, Rows);
+	mywindow(1, 1, Columns, Rows);
+    }
     screen_start();
 }
 
@@ -1664,24 +1731,23 @@ mch_set_winsize(void)
 
 /*
  * call shell, return FAIL for failure, OK otherwise
- * options == SHELL_FILTER if called by do_filter()
- * options == SHELL_COOKED if term needs cooked mode
+ * options: SHELL_*, see vim.h.
  */
     int
 mch_call_shell(
     char_u	*cmd,
     int		options)
 {
-    int	    x;
+    int		x;
 #ifndef DJGPP
-    char_u  *newcmd;
+    char_u	*newcmd;
 #endif
 
     out_flush();
 
     if (options & SHELL_COOKED)
 	settmode(TMODE_COOK);	/* set to normal mode */
-    set_interrupts(FALSE);	    /* restore interrupts */
+    set_interrupts(FALSE);	/* restore interrupts */
 
 #ifdef DJGPP
     /* ignore signals while external command is running */
@@ -1722,7 +1788,7 @@ mch_call_shell(
     settmode(TMODE_RAW);	/* set to raw mode */
     set_interrupts(TRUE);	/* catch interrupts */
 
-    if (x && !expand_interactively)
+    if (x && !(options & SHELL_SILENT))
     {
 	msg_putchar('\n');
 	msg_outnum((long)x);
@@ -1793,7 +1859,7 @@ dos_expandpath(
     struct growarray	*gap,
     char_u		*path,
     char_u		*wildc,
-    int			flags)
+    int			flags)		/* EW_* flags */
 {
     char_u		*buf;
     char_u		*p, *s, *e;
@@ -1834,7 +1900,7 @@ dos_expandpath(
     *e = NUL;
     /* now we have one wildcard component between s and e */
 
-    /* if the file name ends in "*" and does not contain a ".", addd ".*" */
+    /* if the file name ends in "*" and does not contain a ".", add ".*" */
     if (e[-1] == '*' && vim_strchr(s, '.') == NULL)
     {
 	*e++ = '.';
@@ -1890,7 +1956,7 @@ dos_expandpath(
 mch_expandpath(
     struct growarray	*gap,
     char_u		*path,
-    int			flags)
+    int			flags)		/* EW_* flags */
 {
     return dos_expandpath(gap, path, path, flags);
 }
@@ -1969,6 +2035,11 @@ mch_rename(const char *OldFile, const char *NewFile)
 	retval = rename(TempFile, NewFile);
 	close(fd);
 	mch_remove((char_u *)OldFile);
+
+	/* If renaming to NewFile failed, rename TempFile back to OldFile, so
+	 * that it looks like nothing happened. */
+	if (retval)
+	    rename(TempFile, OldFile);
     }
     vim_free(TempFile);
 
