@@ -156,6 +156,16 @@
 # define W_OK 2			/* for systems that don't have W_OK in unistd.h */
 #endif
 
+/*
+ * System definitions
+ */
+
+/* Open VMS is case insensitive */
+#define CASE_INSENSITIVE_FILENAME
+
+/* Use binary fileio */
+#define BINARY_FILE_IO
+
 /* #define USE_TMPNAM */		/* use tmpnam() instead of mktemp() */
 
 /*
@@ -212,15 +222,7 @@
 #endif
 
 #ifndef VIM_HLP
-/*
- * If the line with $VIMRUNTIME doesn't work for you (":help" doesn't find the
- * help files), then change the zero into a one and try again.
- */
-# if 0
-#  define VIM_HLP		"$VIM_DOC:help.txt"
-# else
-#  define VIM_HLP		"$VIMRUNTIME/doc/help.txt"
-# endif
+# define VIM_HLP		"$VIMRUNTIME/doc/help.txt"
 #endif
 
 #ifndef SYNTAX_FNAME
@@ -234,11 +236,11 @@
 #endif /* VIMINFO */
 
 #ifndef DEF_BDIR
-# define DEF_BDIR		"./,sys$disk:[],tmp:,sys$login:"	/* default for 'backupdir' */
+# define DEF_BDIR		"./,tmp:,sys$login:"	/* default for 'backupdir' */
 #endif
 
 #ifndef DEF_DIR
-# define DEF_DIR		"./,sys$disk:[],tmp:,sys$login:"	/* default for 'directory' */
+# define DEF_DIR		"./,tmp:,sys$login:"	/* default for 'directory' */
 #endif
 
 #define TEMPNAME		"tmp:v?XXXXXX.txt"
